/* Copyright (c) 2001-2022, David A. Clunie DBA Pixelmed Publishing. All rights reserved. */

package com.pixelmed.convert;

import com.pixelmed.dicom.Attribute;
import com.pixelmed.dicom.AttributeList;
import com.pixelmed.dicom.CodedSequenceItem;
import com.pixelmed.dicom.CodeStringAttribute;
import com.pixelmed.dicom.DicomDictionary;
import com.pixelmed.dicom.DicomException;
import com.pixelmed.dicom.SequenceAttribute;
import com.pixelmed.dicom.TagFromName;
import com.pixelmed.dicom.UnlimitedTextAttribute;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import com.pixelmed.slf4j.Logger;
import com.pixelmed.slf4j.LoggerFactory;

/**
 * <p>A class for encoding channel parameters describing immunostaining (including antibodies, fluorescence).</p>
 *
 * @author	dclunie
 */

public class Immunostaining {
	private static final String identString = "@(#) $Header: /userland/cvs/pixelmed/imgbook/com/pixelmed/convert/Immunostaining.java,v 1.3 2022/03/06 21:45:13 dclunie Exp $";

	private static final Logger slf4jlogger = LoggerFactory.getLogger(Immunostaining.class);
	
	private static Map<String,CodedSequenceItem> codedTargets = new HashMap<String,CodedSequenceItem>();
	
	static {
		try {
			// matcher will capitalize key first
			codedTargets.put("NUCLEI",new CodedSequenceItem("84640000","SCT","Nucleus"));
			
			codedTargets.put("CD3",new CodedSequenceItem("44706009","SCT","CD3"));
			codedTargets.put("CD4",new CodedSequenceItem("24655002","SCT","CD4"));
			codedTargets.put("CD8",new CodedSequenceItem("54237000","SCT","CD8"));
			codedTargets.put("CD11B",new CodedSequenceItem("27130004","SCT","CD11b"));
			codedTargets.put("CD11C",new CodedSequenceItem("46959001","SCT","CD11c"));
			codedTargets.put("CD20",new CodedSequenceItem("82753007","SCT","CD20"));
			codedTargets.put("CD45",new CodedSequenceItem("19677004","SCT","CD45"));
			codedTargets.put("CD56",new CodedSequenceItem("42891003","SCT","CD56"));
			codedTargets.put("CD66B",new CodedSequenceItem("8164002","SCT","CD66B"));	// apparently the same as CD67 :(
			codedTargets.put("CD68",new CodedSequenceItem("31001006","SCT","CD68"));
			codedTargets.put("CD163",new CodedSequenceItem("C104064","NCIt","CD163"));
			codedTargets.put("CD169",new CodedSequenceItem("C0142251","UMLS","CD169"));

			codedTargets.put("CCR2",new CodedSequenceItem("C101577","NCIt","CCR2"));
			codedTargets.put("EOMES",new CodedSequenceItem("C102928","NCIt","EOMES"));
			codedTargets.put("FOXP3",new CodedSequenceItem("C104394","NCIt","FOXP3"));
			codedTargets.put("GATA3",new CodedSequenceItem("C75494","NCIt","GATA3"));
			codedTargets.put("GRZB",new CodedSequenceItem("130636000","SCT"," Granzyme B"));
			codedTargets.put("GRANZYME B",new CodedSequenceItem("130636000","SCT"," Granzyme B"));
			codedTargets.put("IDO",new CodedSequenceItem("C79782","NCIt","IDO"));
			codedTargets.put("KI67",new CodedSequenceItem("C0208804","UMLS","KI67"));
			codedTargets.put("NKP46",new CodedSequenceItem("C129059","NCIt","NKP46"));	// also NCR1
			codedTargets.put("RORGT",new CodedSequenceItem("C106314","NCIt","RORGT"));
			codedTargets.put("RORGAMMAT",new CodedSequenceItem("C106314","NCIt","RORGT"));
			codedTargets.put("TBET",new CodedSequenceItem("C104249","NCIt","TBET"));	// also TBX21

			codedTargets.put("PD1",new CodedSequenceItem("C94697","NCIt","PD1"));
			codedTargets.put("PD-1",new CodedSequenceItem("C94697","NCIt","PD1"));
			codedTargets.put("PDL1",new CodedSequenceItem("C96024","NCIt","PDL1"));	// Programmed Cell Death 1 Ligand 1 NCIt C96024 (is also CD274)
			codedTargets.put("PD-L1",new CodedSequenceItem("C96024","NCIt","PDL1"));
			
			codedTargets.put("HLAI",new CodedSequenceItem("C20706","NCIt","HLA Class I"));
			codedTargets.put("HLA I",new CodedSequenceItem("C20706","NCIt","HLA Class I"));
			codedTargets.put("HLAII",new CodedSequenceItem("C20705","NCIt","HLA Class II"));
			codedTargets.put("HLA II",new CodedSequenceItem("C20705","NCIt","HLA Class II"));
			codedTargets.put("HLA DRB1",new CodedSequenceItem("C52341","NCIt","HLA DRB1"));
		}
		catch (DicomException e) {
			// ignore
		}
	}
	
	public class ImmunostainingChannel {
		//Channel ID	Channel Name	Cycle Number	Sub-Cycle #	Antibody name	Clone	Vendor	Catalog #
		String channelID;
		String channelName;
		String cycleNumber;
		String subCycleNumber;
		String antibodyName;
		String clone;
		String vendor;
		String catalogNumber;
		
		public ImmunostainingChannel(String channelID,String channelName,String cycleNumber,String subCycleNumber,String antibodyName,String clone,String vendor,String catalogNumber) {
			this.channelID = channelID;
			this.channelName = channelName;
			this.cycleNumber = cycleNumber;
			this.subCycleNumber = subCycleNumber;
			this.antibodyName = antibodyName;
			this.clone = clone;
			this.vendor = vendor;
			this.catalogNumber = catalogNumber;
		}
		
		public String toString() {
			StringBuffer buf = new StringBuffer();
			buf.append("channelID = "+channelID);
			buf.append(", channelName = "+channelName);
			buf.append(", cycleNumber = "+cycleNumber);
			buf.append(", subCycleNumber = "+subCycleNumber);
			buf.append(", antibodyName = "+antibodyName);
			buf.append(", clone = "+clone);
			buf.append(", vendor = "+vendor);
			buf.append(", catalogNumber = "+catalogNumber);
			return buf.toString();
		}
		
		SequenceAttribute getSpecimenPreparationStepContentItemSequence() throws DicomException {
			SequenceAttribute specimenPreparationStepContentItemSequence = new SequenceAttribute(DicomDictionary.StandardDictionary.getTagFromName("SpecimenPreparationStepContentItemSequence"));
			{
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"255592004","SCT","Channel");
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(channelID); itemList.put(a); }
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			// put channelName both as TEXT and CODE (if recognized), since may need literal match on channelName string if present in out-of-band metadata
			{
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"C25702","NCIt","Target");	// hmmm - very generic, but "Antigen" is too specific :(
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(channelName); itemList.put(a); }
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			if (channelName != null && channelName.length() > 0) {
				AttributeList itemList = new AttributeList();
				CodedSequenceItem targetCode = codedTargets.get(channelName.toUpperCase().trim());
				if (targetCode != null) {
					slf4jlogger.debug("getSpecimenPreparationStepContentItemSequence(): have code {} for {}",targetCode,channelName);
					{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("CODE"); itemList.put(a); }
					CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"C25702","NCIt","Target");	// hmmm - very generic, but "Antigen" is too specific :(
					CodedSequenceItem.putSingleCodedSequenceAttribute(itemList,TagFromName.ConceptCodeSequence,targetCode);
					specimenPreparationStepContentItemSequence.addItem(itemList);
				}
				else {
					slf4jlogger.debug("getSpecimenPreparationStepContentItemSequence(): cannot find code for {}",channelName);
				}
			}
			else {
				slf4jlogger.error("getSpecimenPreparationStepContentItemSequence(): missing channelName for channelID {}",channelID);
			}
			if (cycleNumber != null && cycleNumber.length() > 0) {
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"C25472","NCIt","Cycle");
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(cycleNumber); itemList.put(a); }
				if (subCycleNumber != null && subCycleNumber.length() > 0) {
					SequenceAttribute modifier = new SequenceAttribute(TagFromName.ContentItemModifierSequence);
					AttributeList modifierList = new AttributeList();
					{
						{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); modifierList.put(a); }
						CodedSequenceItem.putSingleCodedSequenceItem(modifierList,TagFromName.ConceptNameCodeSequence,"C25472","NCIt","Cycle");
						{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(subCycleNumber); modifierList.put(a); }
					}
					modifier.addItem(modifierList);
					itemList.put(modifier);
				}
				else {
					slf4jlogger.info("getSpecimenPreparationStepContentItemSequence(): missing subCycleNumber for channelID {}",channelID);
				}
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			else {
				slf4jlogger.info("getSpecimenPreparationStepContentItemSequence(): missing cycleNumber for channelID {}",channelID);
			}
			{
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("CODE"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"111701","DCM","Processing type");
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptCodeSequence,"406858009","SCT","Fluorescent staining");
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			if (antibodyName != null && antibodyName.length() > 0) {
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"424361007","SCT","Using substance");
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(antibodyName); itemList.put(a); }
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			else {
				slf4jlogger.info("getSpecimenPreparationStepContentItemSequence(): missing antibodyName for channelID {}",channelID);
			}
			if (clone != null && clone.length() > 0) {
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"C37925","NCIt","Clone");		// this is generic "A group of genetically identical cells, organisms, or DNA molecules" - would prefer more specific "antibody clone" :(
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(clone); itemList.put(a); }
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			else {
				slf4jlogger.info("getSpecimenPreparationStepContentItemSequence(): missing clone for channelID {}",channelID);
			}
			if (vendor != null && vendor.length() > 0) {
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"C0947322","UMLS","Manufacturer Name");
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(vendor); itemList.put(a); }
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			else {
				slf4jlogger.info("getSpecimenPreparationStepContentItemSequence(): missing vendor for channelID {}",channelID);
			}
			if (catalogNumber != null && catalogNumber.length() > 0) {
				AttributeList itemList = new AttributeList();
				{ Attribute a = new CodeStringAttribute(TagFromName.ValueType); a.addValue("TEXT"); itemList.put(a); }
				CodedSequenceItem.putSingleCodedSequenceItem(itemList,TagFromName.ConceptNameCodeSequence,"111529","DCM","Brand Name");
				{ Attribute a = new UnlimitedTextAttribute(TagFromName.TextValue); a.addValue(catalogNumber); itemList.put(a); }
				specimenPreparationStepContentItemSequence.addItem(itemList);
			}
			else {
				slf4jlogger.info("getSpecimenPreparationStepContentItemSequence(): missing catalogNumber for channelID {}",channelID);
			}

			return specimenPreparationStepContentItemSequence;
		}
	}
	
	SortedMap<String,ImmunostainingChannel> channelsByChannelID = null;
	SortedMap<String,SequenceAttribute> specimenPreparationStepContentItemSequenceByChannelID = null;

	public SortedMap<String,SequenceAttribute> getMapOfSpecimenPreparationStepContentItemSequenceByChannelID() throws DicomException {
		if (specimenPreparationStepContentItemSequenceByChannelID == null) {	// lazy instantiation
			specimenPreparationStepContentItemSequenceByChannelID = new TreeMap<String,SequenceAttribute>();
			for (ImmunostainingChannel channel : channelsByChannelID.values()) {
				specimenPreparationStepContentItemSequenceByChannelID.put(channel.channelID,channel.getSpecimenPreparationStepContentItemSequence());
			}
		}
		return specimenPreparationStepContentItemSequenceByChannelID;
	}
	
	private String getColumnValueIfPresent(String columnName,Map<String,Integer> columnIndexByColumnHeader,String[] values) {
		Integer columnIndex = columnIndexByColumnHeader.get(columnName);
		return columnIndex == null ? null : values[columnIndex.intValue()];
	}
	
	private void readFile(File file) throws FileNotFoundException, IOException, DicomException {
		Map<String,Integer> columnIndexByColumnHeader = null;
		BufferedReader r = new BufferedReader(new FileReader(file));
		String line = null;
		while ((line=r.readLine()) != null) {
			if (columnIndexByColumnHeader == null) {
				String[] columnHeaders = null;
				if (line.contains("\t")) {
					slf4jlogger.debug("readFile(): have tab separated values");
					columnHeaders = line.split("\t");
				}
				else if (line.contains(",")) {
					slf4jlogger.debug("readFile(): have comma separated values");
					columnHeaders = line.split(",");
				}
				if (columnHeaders != null) {
					columnIndexByColumnHeader = new HashMap<String,Integer>();
					int columnIndex=0;
					for (String columnHeader : columnHeaders) {
						columnIndexByColumnHeader.put(columnHeader,new Integer(columnIndex));
						++columnIndex;
					}
				}
				else {
					throw new DicomException("Unrecognized form of header row in input file");
				}
			}
			else {
				// process actual row of data
				String[] values = null;
				if (line.contains("\t")) {
					values = line.split("\t");
				}
				else if (line.contains(",")) {
					values = line.split(",");
				}
				if (values != null) {
					//Channel ID	Channel Name	Cycle Number	Sub-Cycle #	Antibody name	Clone	Vendor	Catalog #
					//Channel ID,Channel Name,Passed QC
					String channelID = getColumnValueIfPresent("Channel ID",columnIndexByColumnHeader,values);
					if (channelID == null || channelID.length() == 0) {
						throw new DicomException("Missing Channel ID");
					}
					else {
						// 1
						// Channel:0:11
						channelID=channelID.replace("Channel:0:","");	// need to be consistent about this when matching against any ID from OME-TIFF-XML (001323)
						String channelName = getColumnValueIfPresent("Channel Name",columnIndexByColumnHeader,values);
						String cycleNumber = getColumnValueIfPresent("Cycle Number",columnIndexByColumnHeader,values);
						String subCycleNumber = getColumnValueIfPresent("Sub-Cycle #",columnIndexByColumnHeader,values);
						String antibodyName = getColumnValueIfPresent("Antibody name",columnIndexByColumnHeader,values);
						String clone = getColumnValueIfPresent("Clone",columnIndexByColumnHeader,values);
						String vendor = getColumnValueIfPresent("Vendor",columnIndexByColumnHeader,values);
						String catalogNumber = getColumnValueIfPresent("Catalog #",columnIndexByColumnHeader,values);
					
						ImmunostainingChannel channel = new ImmunostainingChannel(channelID,channelName,cycleNumber,subCycleNumber,antibodyName,clone,vendor,catalogNumber);
						if (slf4jlogger.isDebugEnabled()) slf4jlogger.debug("readFile(): have channel {}",channel.toString());
						channelsByChannelID.put(channelID,channel);
					}
				}
				else {
					throw new DicomException("Unrecognized form of data row in input file "+file);
				}

			}
		}
		r.close();
	}
	
	public Immunostaining(File file) throws FileNotFoundException, IOException, DicomException {
		channelsByChannelID = new TreeMap<String,ImmunostainingChannel>();
		readFile(file);
	}

	/**
	 * <p>Read a TIFF image input format file consisting of one or more pages or tiles, and create one or more images of a specified or appropriate SOP Class.</p>
	 */
	public static void main(String arg[]) {
		try {
			new Immunostaining(new File(arg[0]));
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
}

